local _M = {}


local how_do_i_die = {}

function _M.AddDyingCallback(fn)
	table.insert(how_do_i_die, fn)
end


local function new_handler_table()
	return {
		is_valid = {},
	}
end

_M.handlers = new_handler_table()
_M.handlers.RESURRECT = new_handler_table()


--[[
-- Top level handlers.
--]]

local function DoDie(player, data)
	for _, fn in ipairs(how_do_i_die) do
		fn(player, data)
	end
end
_M.DoDie = DoDie
_M.handlers.DIE = DoDie

function _M.handlers.RELOAD(player, data)
	return _G.StartNextInstance({
		reset_action = _G.RESET_ACTION.LOAD_SLOT,
		save_slot = _G.SaveGameIndex:GetCurrentSaveSlot(),
	})
end

function _M.handlers.PROTO_RESURRECT(player, data)
	return use("dialogs").Resurrect(player, data)
end



--[[
-- General resurrection utilities.
--]]

local function TurnIntoDummyChildEntity(inst, parent, remove_cb)
	inst.persists = false
	if not inst.Transform then
		inst.entity:AddTransform()
	end
	inst:AddTag("NOCLICK")
	inst:AddTag("NOBLOCK")

	if inst.parent ~= parent then
		parent:AddChild(inst)
	end
	_G.assert( inst.parent == parent )

	inst.Transform:SetPosition( 0, 0, 0 )

	inst:ListenForEvent("onremove", function()
		if remove_cb then
			remove_cb(inst)
		end
		inst:Remove()
	end, parent)
	inst:ListenForEvent("onremove", function(inst)
		if parent:IsValid() and inst.parent == parent then
			parent:RemoveChild(inst)
		end
	end)

	return inst
end

local function MakeDummyChildEntity(parent, remove_cb)
	return TurnIntoDummyChildEntity(_G.CreateEntity(), parent, remove_cb)
end

local function MakeWarnerEntity(parent)
	local inst = MakeDummyChildEntity(parent)

	local l = inst.entity:AddLabel()
	l:SetFontSize(30)
	l:SetFont(_G.DEFAULTFONT)
	l:SetPos(0, 4, 0)
	l:SetColour(.7, 0, 0)
	l:Enable(false)

	function inst:Say(line)
		self.Label:SetText(line)
		self.Label:Enable(true)
	end

	function inst:ShutUp()
		self.Label:SetText("")
		self.Label:Enable(false)
	end

	return inst
end

local GetWarnerEntity = (function()
	local warners = _G.setmetatable({}, {__mode = "kv"})

	return function(parent)
		_G.assert( parent:IsValid() )

		local w = warners[parent]
		if not w then
			w = MakeWarnerEntity(parent)
			warners[parent] = w
		end
		return w
	end
end)()

AddStategraphState("wilson", _G.State {
	name = _M,
	tags = {"busy"},

	onenter = function(inst, params)
		inst.components.playercontroller:Enable(false)

		inst.AnimState:PlayAnimation(params.anim)
		if params.sound then
			inst.SoundEmitter:PlaySound(params.sound)
		end

		_G.TheCamera:SetDistance(14)
		if inst.HUD then
			inst.HUD:Hide()
		end
	end,

	onexit = function(inst)
		if inst.HUD then
			inst.HUD:Show()
		end
		_G.TheCamera:SetDefault()
		inst.components.playercontroller:Enable(true)
	end,

	events = {
		_G.EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
	},
})

local function add_infinite_resurrector(inst)
	inst:AddComponent("resurrector")

	local res = inst.components.resurrector
	res.Resurrect = (function()
		local oldResurrect = res.Resurrect
		return function(self, ...)
			oldResurrect(self, ...)
			self.used = false
		end
	end)()

	return res
end

local function basic_resurrect_for_state(anim, sound, skip_day)
	return function(inst, dude)
		if dude.components.talker then
			dude.components.talker.ignoring = true
		end

		dude:DoTaskInTime(0, function(dude)
			if _G.GetClock():IsNight() or not dude.LightWatcher:IsInLight() then
				TurnIntoDummyChildEntity(_G.SpawnPrefab("exitcavelight"), dude)
			end

			if dude.components.talker then
				dude.components.talker.ignoring = nil
			end
		end)

		dude.sg:GoToState(_M, {anim = anim, sound = sound})

		dude.components.health:SetInvincible(false)
		dude.components.health:Respawn( _G.TUNING.RESURRECT_HEALTH )
		dude.components.hunger:SetPercent( math.max(1/2, math.min(2/3, dude.components.hunger:GetPercent())) )
		dude.components.sanity:SetPercent( math.max(1/2, math.min(1/2, dude.components.sanity:GetPercent())) )
		dude.components.health:SetInvincible(true)

		if skip_day then
			_G.GetClock():MakeNextDay()
		end

			
		return dude:StartThread(function()
			local time_remaining = 10
			local update_period = 0.5

			local warner = GetWarnerEntity(dude)

			repeat
				warner:Say(("Invulnerable for %.1f seconds"):format(time_remaining))
				_G.Sleep(update_period)
				time_remaining = time_remaining - update_period
			until time_remaining <= 0

			dude.components.health:SetInvincible(false)
			warner:ShutUp()
		end)
	end
end


--[[
-- Personal resurrector.
--]]

local function MakePersonalResurrector()
	local personal_resurrector

	AddSimPostInit(function(player)
		personal_resurrector = MakeDummyChildEntity(player)

		local res = add_infinite_resurrector(personal_resurrector)
		res.active = false
		res.doresurrect = basic_resurrect_for_state("amulet_rebirth", "dontstarve/common/rebirth_amulet_raise", false)
	end)


	local function is_handler_valid(player, data)
		return personal_resurrector and personal_resurrector:IsValid() and personal_resurrector.components.resurrector
	end

	local function resurrection_handler(player, data)
		personal_resurrector.components.resurrector.active = true
		return DoDie(player, data)
	end

	return resurrection_handler, is_handler_valid
end

_M.handlers.RESURRECT.HERE, _M.handlers.RESURRECT.is_valid.HERE = MakePersonalResurrector()


--[[
-- Prefab (e.g., camp) resurrectors.
--]]

local function MakePrefabResurrector(prefab)
	local camp_resurrectors = _G.setmetatable({}, {__mode = "k"})

	local function is_valid_resurrector(inst)
		return inst:IsValid() and inst.parent:IsValid() and inst.components.resurrector
	end

	local function is_acceptable_resurrector(inst)
		return is_valid_resurrector(inst) and not inst.parent:IsInLimbo()	
	end

	local function purge_resurrector(inst)
		camp_resurrectors[inst] = nil
		inst:Remove()
	end

	local function count_resurrectors()
		local cnt = 0

		local to_remove = {}
		for inst in pairs(camp_resurrectors) do
			if is_acceptable_resurrector(inst) then
				cnt = cnt + 1
			elseif not is_valid_resurrector(inst) then
				table.insert(to_remove, inst)
			end
		end
		for _, inst in ipairs(to_remove) do
			purge_resurrector(inst)
		end

		return cnt
	end

	local function toggle_camps_activity(b)
		for inst in pairs(camp_resurrectors) do
			if inst.components.resurrector then
				inst.components.resurrector.active = b and is_acceptable_resurrector(inst)
			end
		end
	end

	local function do_camp_resurrect(inst, dude)
		_G.TheFrontEnd:Fade(true, 3)

		local offset = _G.FindWalkableOffset(
			inst:GetPosition(),
			2*math.pi*math.random(),
			math.max(2, inst.parent.Physics and inst.parent.Physics:GetRadius() or 0) + 1,
			32,
			true,
			false
		) or _G.Point()

		dude.Physics:Teleport( (inst:GetPosition() + offset):Get() )

		basic_resurrect_for_state("wakeup", nil, true)(inst, dude)

		toggle_camps_activity(false)
	end

	AddPrefabPostInit(prefab, function(camp)
		local inst = MakeDummyChildEntity(camp, purge_resurrector)

		local res = add_infinite_resurrector(inst)
		res.active = false
		res.doresurrect = do_camp_resurrect

		camp_resurrectors[inst] = true
	end)


	local function resurrection_handler(player, data)
		toggle_camps_activity(true)
		return DoDie(player, data)
	end

	local function is_handler_valid(player, data)
		return count_resurrectors() > 0
	end

	return resurrection_handler, is_handler_valid
end

_M.handlers.RESURRECT.CAMP, _M.handlers.RESURRECT.is_valid.CAMP = MakePrefabResurrector "firepit"


--[[
-- "Back" button.
--]]

function _M.handlers.RESURRECT.NEVERMIND(player, data)
	return use("dialogs").Main(player, data)
end


return _M
